                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _BSP_H_
#define _BSP_H_

class BSPTree;

enum PASS {
	PASS_OPAQUE,
	PASS_TRANSLUCENT
};

#include "OpenGL/Frustum.h"

#include "OpenGL/Light.h"
#include "Common/Set.h"
#include "Polygon.h"

Frustum frustum;

/* ########################################################################################################### */

class BSPNode {
public:
	BSPNode(){
		polygon = NULL;
		back = front = NULL;
	}

	~BSPNode(){
		if (polygon != NULL) delete polygon;
		if (back  != NULL) delete back;
		if (front != NULL) delete front;
	}

	class Polygon *polygon;

	BSPNode *front;
	BSPNode *back;

	void draw(PASS pass, Vertex &pos, TextureHandler *texHandler, Set <Light *> &lights);
	void drawFogDepthAlpha(Vertex &pos, Vertex &fogPos, float fogRadius, float constDist);

	void build(Set <class Polygon *> &polygons);
	bool lineOfSight(Vertex &v0, Vertex &v1, Vertex &result, class Polygon *&poly);
	bool push(Vertex &pos, Vertex &oldPos, float distance);

	int getSize(){
		int res = 1;
		if (back  != NULL) res += back->getSize();
		if (front != NULL) res += front->getSize();
		return res;
	}

};


/* ---------------------------------------------------------- */

void BSPNode::draw(PASS pass, Vertex &pos, TextureHandler *texHandler, Set <Light *> &lights){
	int polyFlags = polygon->getFlags();
	RELATION rel = polygon->vertexRelation(pos);

    int frustumRel = frustum.planeIntersectsFrustum(polygon->getNormal(), polygon->getOffset());

    if (pass == PASS_OPAQUE){
		if (rel == BACK){
            if (frustumRel <= 0){
    			if (back != NULL) back->draw(pass, pos, texHandler, lights);
            }
            if (frustumRel == 0){
                if (polygon->isInFrustum(frustum)){
    			    if ((polyFlags & PF_DOUBLESIDED) != 0 && (polyFlags & PF_TRANSLUCENT) == 0) polygon->draw(pass, pos, texHandler, lights);
                }
            }
            if (frustumRel >= 0){
    			if (front != NULL) front->draw(pass, pos, texHandler, lights);
            }
		} else {
            if (frustumRel >= 0){
			    if (front != NULL) front->draw(pass, pos, texHandler, lights);
            }
            if (frustumRel == 0){
                if (polygon->isInFrustum(frustum)){
                    if ((polyFlags & PF_TRANSLUCENT) == 0) polygon->draw(pass, pos, texHandler, lights);
                }
            }
            if (frustumRel <= 0){
			    if (back != NULL) back->draw(pass, pos, texHandler, lights);
            }
		}
	} else if (pass == PASS_TRANSLUCENT){
		if (rel == BACK){
            if (frustumRel >= 0){
    			if (front != NULL) front->draw(pass, pos, texHandler, lights);
            }
            if (frustumRel == 0){
                if (polygon->isInFrustum(frustum)){
        			if ((polyFlags & PF_DOUBLESIDED) != 0 && (polyFlags & PF_TRANSLUCENT) != 0) polygon->draw(pass, pos, texHandler, lights);
                }
            }
            if (frustumRel <= 0){
    			if (back != NULL) back->draw(pass, pos, texHandler, lights);
            }
		} else {
            if (frustumRel <= 0){
    			if (back != NULL) back->draw(pass, pos, texHandler, lights);
            }
            if (frustumRel == 0){
                if (polygon->isInFrustum(frustum)){
    			    if ((polyFlags & PF_TRANSLUCENT) != 0) polygon->draw(pass, pos, texHandler, lights);
                }
            }
            if (frustumRel >= 0){
			    if (front != NULL) front->draw(pass, pos, texHandler, lights);
            }
		}
	}

}


void BSPNode::drawFogDepthAlpha(Vertex &pos, Vertex &fogPos, float fogRadius, float constDist){
	//int polyFlags = polygon->getFlags();
	RELATION rel = polygon->vertexRelation(pos);

	float dist = polygon->distance(fogPos);

	if (fabs(dist) < fogRadius){
		if (polygon->sphereIntersects(fogPos, fogRadius)){
			polygon->drawFogDepthAlpha(pos, constDist);

		}
	} 

	if (back != NULL) back->drawFogDepthAlpha(pos, fogPos, fogRadius, constDist);
	if (front != NULL) front->drawFogDepthAlpha(pos, fogPos, fogRadius, constDist);


}


void BSPNode::build(Set <class Polygon *> &polygons){
	Set <class Polygon *> backList, frontList;

	int cutterScore = 0x7FFFFFFF, score, diff;
	int i, j, rel, cutterIndex = 0;
	
	
	// Select the cutting surface
	for (i = 0; i < polygons.getSize(); i++){
		score = 0;
		diff = 0;
		for (j = 0; j < polygons.getSize(); j++){
			if (i != j){
				rel = polygons[i]->polygonRelation(*polygons[j]);

				if (rel == CUTS){
					score += 3;
				} else {
					if (rel == FRONT) diff++; else diff--;
				}
			}
		}
		score += abs(diff);
		
		if (score < cutterScore){
			cutterIndex = i;
			cutterScore = score;
		}
	}

	polygon = polygons[cutterIndex];
	polygons.remove(cutterIndex);
	
	// Split with the selected surface
	class Polygon *backPoly, *frontPoly;
	for (i = 0; i < polygons.getSize(); i++){
		rel = polygon->split(*polygons[i], backPoly, frontPoly);

		if (rel == BACK)  backList.add(polygons[i]);
		if (rel == FRONT || rel == PLANAR) frontList.add(polygons[i]);
		if (rel == CUTS){
			backList.add(backPoly);
			frontList.add(frontPoly);
			delete polygons[i];
		}
	}

	if (backList.getSize() > 0){
		back = new BSPNode;
		back->build(backList);
	} else back = NULL;

	if (frontList.getSize() > 0){
		front = new BSPNode;
		front->build(frontList);
	} else front = NULL;

}

bool BSPNode::lineOfSight(Vertex &v0, Vertex &v1, Vertex &result, class Polygon *&poly){
	RELATION vr0, vr1;

	vr0 = polygon->vertexRelation(v0);
	vr1 = polygon->vertexRelation(v1);

	if (vr0 == FRONT){
		if (vr1 >= PLANAR){
			if (front != NULL) return front->lineOfSight(v0,v1, result, poly);
		} else {
			if (front != NULL){
				if (front->lineOfSight(v0,v1, result, poly)) return true;
			}
			if (((polygon->getFlags() & PF_NONBLOCKING) == 0) && polygon->linePassesThrough(v0,v1, result)){
				poly = polygon;
				return true;
			}
			if (back != NULL){
				if (back->lineOfSight(v0,v1, result, poly)) return true;
			}
		}
	} else {
		if (vr1 == FRONT){
			if (back != NULL){
				if (back->lineOfSight(v0,v1, result, poly)) return true;
			}
			if ((polygon->getFlags() & PF_DOUBLESIDED) && ((polygon->getFlags() & PF_NONBLOCKING) == 0)){
				if (polygon->linePassesThrough(v0,v1, result)){
					poly = polygon;
					return true;
				}
			}
			if (front != NULL){
				if (front->lineOfSight(v0,v1, result, poly)) return true;
			}
		} else {
			if (back != NULL) return back->lineOfSight(v0,v1, result, poly);
		}
	}

	return false;
}

bool BSPNode::push(Vertex &pos, Vertex &oldPos, float distance){
	float d = polygon->distance(pos);
	RELATION s0 = (d < 0.0f)? BACK : FRONT;
	RELATION rel = polygon->vertexRelation(oldPos);
    int flags = polygon->getFlags();

	d = fabsf(d);
	Vertex hitPoint, normal;
	normal = polygon->getNormal();

	if (((flags & PF_NONBLOCKING) == 0) && polygon->linePassesThrough(oldPos, pos, hitPoint)){
        if (flags & PF_DOUBLESIDED){
            if (rel == s0){
                pos += (distance - d + 0.05f) * rel * normal;
            } else {
                pos += (distance + d + 0.05f) * rel * normal;
            }
        } else {
            pos += (distance + d + 0.05f) * normal;
        }
        return true;
    }

	if (d < distance){
		if (((flags & PF_NONBLOCKING) == 0) && polygon->linePassesThrough(pos, pos - 2 * d * rel * normal, hitPoint)){
			if (flags & PF_DOUBLESIDED){
				if (rel == s0){
					pos += (distance - d + 0.05f) * rel * normal;
				} else {
					pos += (distance + d + 0.05f) * rel * normal;
				}
			} else {
				pos += (distance - d + 0.05f) * normal;
			}
			return true;
		}

		if (front != NULL){
			if (front->push(pos, oldPos, distance)) return true;
		}
		if (back != NULL){
			return back->push(pos, oldPos, distance);
		}
	} else {
		if (rel == FRONT){
			if (front != NULL) return front->push(pos, oldPos, distance);
		} else {
			if (back  != NULL) return back->push(pos, oldPos, distance);
		}
	}

	return false;
}

/* ########################################################################################################### */

class BSPTree {
private:
	BSPNode *top;

public:

	BSPTree(Set <class Polygon *> &polygons);
	~BSPTree();

	void draw(PASS pass, Vertex &pos, TextureHandler *texHandler, Set <Light *> &lights);
	void drawFogDepthAlpha(Vertex &pos, Vertex &fogPos, float fogRadius, float constDist);

	bool lineOfSight(Vertex &v0, Vertex &v1, Vertex &result, class Polygon *&poly);
	void push(Vertex &pos, Vertex &oldPos, float distance);

	int getSize(){
		return top->getSize();
	}
};

/* ---------------------------------------------------------- */

BSPTree::BSPTree(Set <class Polygon *> &polygons){
	top = new BSPNode;

	top->build(polygons);
}

BSPTree::~BSPTree(){
	delete top;
}

void BSPTree::draw(PASS pass, Vertex &pos, TextureHandler *texHandler, Set <Light *> &lights){
	top->draw(pass, pos, texHandler, lights);
}

void BSPTree::drawFogDepthAlpha(Vertex &pos, Vertex &fogPos, float fogRadius, float constDist){
	top->drawFogDepthAlpha(pos, fogPos, fogRadius, constDist);
}


bool BSPTree::lineOfSight(Vertex &v0, Vertex &v1, Vertex &result, class Polygon *&poly){
	return top->lineOfSight(v0, v1, result, poly);
}

void BSPTree::push(Vertex &pos, Vertex &oldPos, float distance){
	int i = 0;
	while (i < 50 && top->push(pos, oldPos, distance)) i++;
}


#include "Polygon.cpp"

/* ########################################################################################################### */

#endif // _BSP_H_
