                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#include <afxwin.h>
#include <atlbase.h>
#include <afxtempl.h>

#include <GL/GL.h>
#include <GL/GLU.h>

#include "3D/Resolution.h"

#include "OpenGL/glExtensions.h"
#include "OpenGL/Texture.h"
#include "BSP.h"

#include "T3dLoader.h"
#include "Common/CPU.h"

#include "3D/DemoLooper.h"
#include "Common/IniReader.h"

//#include "FogVolume.h"

#include "bass.h"


#define VK_ENTER 13

/* ################################################################################################################### */

class OpenGLWindow : public CFrameWnd {
public:
    OpenGLWindow();

	void DrawFrame();
protected:
	HDC hdc;
	HGLRC hrc;

	CRect windowedRect;
	int Width, Height;
	int Bpp;
	int Refresh;
	bool fullScreen;
	bool captureMouse;
	bool windowMoved;
	int middleX, middleY;

	bool Key[256];
	bool leftButton;
	float mouseSensibility;
	bool invertMouse;
	int mouseX, mouseY;

    bool showFPS;
    bool showInfo;
    bool showLights;
	bool coronas;

    bool demoMode;
    bool playMusic;

    __int64 cpuHz;

	Vertex position;
	float Wx, Wy, Wz;

	Set <Light *> lights;
	//Set <FogVolume *> fogVolumes;

	TextureHandler *texHandler;
	TextureID text, lightTex, corona;
	BSPTree *bspTree;

    DemoLooper demoLooper;
    float demoTime;


    CString infoString;

    HMUSIC hMusic;

	BOOL PreCreateWindow(CREATESTRUCT &cs);
	int  OnCreate(LPCREATESTRUCT lpCreateStruct);
	void OnSize(UINT nType, int cx, int cy);
	void OnClose();

	void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	void OnKeyUp  (UINT nChar, UINT nRepCnt, UINT nFlags);
	void OnSysKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);

	void OnLButtonDown(UINT nFlags, CPoint point);
	void OnLButtonUp  (UINT nFlags, CPoint point);
	void OnMouseMove  (UINT nFlags, CPoint point);

	void OnKillFocus(CWnd *pNewWnd);

	DECLARE_MESSAGE_MAP ()
};

BEGIN_MESSAGE_MAP(OpenGLWindow, CFrameWnd)
	
	ON_WM_CREATE ()
	ON_WM_CLOSE ()

	ON_WM_KEYDOWN ()
	ON_WM_KEYUP ()
	ON_WM_SYSKEYDOWN ()

	ON_WM_LBUTTONDOWN ()
	ON_WM_LBUTTONUP ()
	ON_WM_MOUSEMOVE ()

	ON_WM_KILLFOCUS ()

END_MESSAGE_MAP()


/* ################################################################################################################### */

void OpenGLWindow::DrawFrame(){
    static bool first = true;

	int i,len;
    static __int64 prev, curr;
    static float currTime;
    float time, speed;


    prev = curr;
    curr = getCycleNumber();
    if (first){
        prev = curr;
        first = false;
    }
    time = (float) (100000.0 * double(curr - prev) / cpuHz);
    currTime += time;
    speed = time * 0.005f;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(75, (GLdouble) Width / (GLdouble) Height, 1, 32000);


    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glRotatef(Wz, 0.0f, 0.0f, 1.0f);
    glRotatef(Wx, 1.0f, 0.0f, 0.0f);
    glRotatef(Wy, 0.0f, 1.0f, 0.0f);

	glScalef(1,1,-1);

    glTranslatef(-position.x, -position.y, -position.z);

    float matrix[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, matrix);
    Vertex dx(matrix[0], matrix[4], matrix[8]);
    Vertex dy(matrix[1], matrix[5], matrix[9]);
    Vertex dz(matrix[2], matrix[6], matrix[10]);

    frustum.updateFrustum();


    Vertex dir(0,0,0);
    if (demoMode){
        LoopNode node = demoLooper.getNode(demoTime);

		dir.x = node.x - position.x;
		dir.y = node.y - position.y;
		dir.z = node.z - position.z;

		speed = length(dir);
		dir /= speed;

        position.x = node.x;
        position.y = node.y;
        position.z = node.z;
        Wx = node.Wx;
        Wy = node.Wy;
        Wz = node.Wz;
        
        demoTime += time * 0.000005f;        
    } else {

        if (Key[VK_LEFT ]) dir -= dx;
        if (Key[VK_RIGHT]) dir += dx;

        if (Key[VK_UP  ]) dir -= dz;
        if (Key[VK_DOWN]) dir += dz;

        if (Key[VK_SHIFT  ]) dir -= dy;
        if (Key[VK_CONTROL]) dir += dy;

        if (Key['1']) Wz += 1.0f;
        if (Key['2']) Wz -= 1.0f;

        if (Key[VK_ENTER]){
            position = Vertex(0,0,0);
            Wx = 0;
            Wy = 0;
            Wz = 0;
        }

        if (lengthSqr(dir) != 0){
            dir.normalize();
        }


        Vertex oldPos = position;
        position += speed * dir;
        bspTree->push(position, oldPos, 50);
    }


    lights[0]->setPosition(Vertex(250 * cosf(currTime * 0.0000523f), 0, 250 * sinf(currTime * 0.0000267f)));

    for (i = 1; i < 9; i++){
        lights[i]->setSize(125 * (1.5f + cosf(currTime * 0.0001f + i * PI / 4)));
    }

    for (i = 22; i < 26; i++){
        lights[i]->setSize(125 * (1.5f + cosf(currTime * 0.00003f + i * 0.5f - 3)));
    }

    static int last[4];
    for (i = 0; i < 4; i++){
        if (last[i] > 0){
            last[i]++;

            if ((rand() & 127) < last[i]) last[i] = -1;

            lights[27 + i]->setSize(0);
        } else {
            last[i]--;

            if ((rand() & 127) < -last[i]) last[i] = 1;

            lights[27 + i]->setSize(250);
        }
    }

	glClear(GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);


	bspTree->draw(PASS_OPAQUE, position, texHandler, lights);

/*	len = playerHandler->getNumberOfPlayers();
	for (i = 0; i < len; i++){
		player = playerHandler->usePlayer(i);
		if (player != NULL){
			if (player->getID() != playerID){
				player->draw(texHandler);
			}
			playerHandler->stopUsingPlayer(i);
		}
	}
*/


	bspTree->draw(PASS_TRANSLUCENT, position, texHandler, lights);


/*	glEnable(GL_DEPTH_TEST);

	len = fogVolumes.getSize();
	for (i = 0; i < len; i++){
		fogVolumes[i]->draw(bspTree, position, texHandler);
	}

	glDisable(GL_DEPTH_TEST);
*/


	Vertex v,r;
	float size;
	// Draw lights
    if (true/*showLights*/){
        texHandler->setBlending(GL_ONE, GL_ONE);
        texHandler->setTextures(lightTex);
        texHandler->apply();
        
        len = showLights? lights.getSize() : 1;

        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LEQUAL);

        for (i = 0; i < len; i++){
            v    = lights[i]->getPosition();
            size = lights[i]->getSize() * 0.05f;
            glColor3fv((GLfloat *) &lights[i]->getColor());

            glBegin(GL_QUADS);
	            r = v - size * dx - size * dy;
		        glTexCoord2f(0,0);
			    glVertex3fv((GLfloat *) &r);

	            r = v + size * dx - size * dy;
		        glTexCoord2f(1,0);
			    glVertex3fv((GLfloat *) &r);

	            r = v + size * dx + size * dy;
		        glTexCoord2f(1,1);
			    glVertex3fv((GLfloat *) &r);

	            r = v - size * dx + size * dy;
		        glTexCoord2f(0,1);
			    glVertex3fv((GLfloat *) &r);
            glEnd();
        }
    }

    glDisable(GL_DEPTH_TEST);


	if (coronas){
		// Draw coronas
		texHandler->setBlending(GL_ONE, GL_ONE);
		//texHandler->setBlending(GL_ONE_MINUS_DST_COLOR, GL_ONE);
		texHandler->setTextures(corona);
		texHandler->apply();
		
		class Polygon *poly;
		Color col;
		bool visible;
		len = lights.getSize();
		for (i = 0; i < len; i++){
			if (lights[i]->hasCorona()){
				v = lights[i]->getPosition();
				visible = !bspTree->lineOfSight(position, v, r, poly);
				
				lights[i]->setLastTime(visible, currTime);
				
				size = lights[i]->getCoronaSize(position, currTime);
				
				if (size > 0){
					col = lights[i]->getColor();
					if (size < 120) col = col * 0.01f * (size - 20);
					glColor3fv((GLfloat *) &col);
					
					glBegin(GL_QUADS);
						r = v - size * dx - size * dy;
						glTexCoord2f(0,0);
						glVertex3fv((GLfloat *) &r);
					
						r = v + size * dx - size * dy;
						glTexCoord2f(1,0);
						glVertex3fv((GLfloat *) &r);
					
						r = v + size * dx + size * dy;
						glTexCoord2f(1,1);
						glVertex3fv((GLfloat *) &r);
						
						r = v - size * dx + size * dy;
						glTexCoord2f(0,1);
						glVertex3fv((GLfloat *) &r);
					glEnd();
				}
			}
		}
	}

	if (showFPS){
		static float fps[8];
		char str[256];

		for (i = 0; i < 7; i++){
			fps[i] = fps[i+1];
		}

		fps[7] = 100000 / time;
		float sum = 0;
		for (i = 0; i < 8; i++) sum += fps[i];
		
		sprintf(str,"%d", (int) (sum / 8 + 0.5f));

		/*CString string, tmp;
		len = playerHandler->getNumberOfPlayers();
		Player *player;
		for (i = 0; i < len; i++){
			player = playerHandler->usePlayer(i);
			if (player != NULL){
				tmp.Format("%d\n", player->getID());
				string += tmp;
				playerHandler->stopUsingPlayer(i);
			}
		}*/

		texHandler->setTextures(text);
        texHandler->setTexEnvs(TEXENV_REPLACE);
		texHandler->setBlending(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		texHandler->apply();
		
        glDrawText(str, 0.02f, 0.02f, 0.03f, 0.03f);
		/*string += "\n";
		string += str;
		glDrawText((char *) (LPCSTR) string, 0.02f, 0.02f, 0.03f, 0.03f);
		*/
	}

    if (showInfo){
		texHandler->setTextures(text);
        texHandler->setTexEnvs(TEXENV_REPLACE);
		texHandler->setBlending(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		texHandler->apply();
        glDrawText((char *) (LPCTSTR) infoString, 0.02f, 0.71f, 0.018f, 0.020f);
    }

	SwapBuffers(hdc);
}


OpenGLWindow::OpenGLWindow(){
	IniReader ini("demo.ini");

	if (!ini.getInt(Width,   "Width"  )) Width   = 1024;
	if (!ini.getInt(Height,  "Height" )) Height  = 768;
	if (!ini.getInt(Bpp,     "Bpp"    )) Bpp     = 32;
	if (!ini.getInt(Refresh, "Refresh")) Refresh = 85;
	if (!ini.getBool(fullScreen, "StartupFullscreen")) fullScreen = true;

	captureMouse = true;

	if (!ini.getFloat(mouseSensibility, "MouseSensibility")) mouseSensibility = 0.1f;
	if (!ini.getBool(invertMouse, "InvertMouse")) invertMouse = true;

	if (!ini.getBool(demoMode, "DemoMode")) demoMode = true;
    demoTime = 0.0f;

    middleX = Width / 2;
	middleY = Height / 2;
	if (fullScreen) setResolution(Width, Height, Bpp, Refresh);

    SetCursorPos(middleX, middleY);

	
	if (!ini.getBool(showFPS,    "ShowFPS"))    showFPS    = true;
	if (!ini.getBool(showInfo,   "ShowInfo"))   showInfo   = false;
	if (!ini.getBool(showLights, "ShowLights")) showLights = false;
	if (!ini.getBool(coronas,    "Coronas"))    coronas    = true;
	
	if (!ini.getBool(playMusic,  "PlayMusic"))  playMusic  = true;

	position = Vertex(0,0,0);

	Wx = 0;
	Wy = 0;
	Wz = 0;

	ZeroMemory(&Key, sizeof(Key));
	leftButton = false;

	CRect r(0, 0, Width, Height);
	windowedRect = r;

	Create(NULL, "OpenGL", WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS, r);
}

BOOL OpenGLWindow::PreCreateWindow(CREATESTRUCT &cs){
	CFrameWnd::PreCreateWindow(cs);

	cs.dwExStyle &= ~WS_EX_CLIENTEDGE;

	return 1;
}


int OpenGLWindow::OnCreate(LPCREATESTRUCT lpCreateStruct){
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1) return -1;

	ShowWindow(SW_SHOW);
	SetForegroundWindow();
	SetFocus();

	hdc = ::GetDC((HWND) *this);

    static PIXELFORMATDESCRIPTOR pfd = {
        sizeof (PIXELFORMATDESCRIPTOR),
			1,
			PFD_DRAW_TO_WINDOW |
			PFD_SUPPORT_OPENGL |
			PFD_DOUBLEBUFFER,
			PFD_TYPE_RGBA,
			Bpp,
			0, 0, 0, 0, 0, 0,
			0, 0,
			0, 0, 0, 0, 0,
			Bpp == 32? 24:16,
			0,
			0,
			PFD_MAIN_PLANE,
			0,
			0, 0, 0
    };

    int nPixelFormat = ChoosePixelFormat(hdc, &pfd);
    SetPixelFormat(hdc, nPixelFormat, &pfd);

	hrc = wglCreateContext(hdc);
	wglMakeCurrent(hdc,hrc);
	
	ShowCursor(FALSE);

    initCPU();

    cpuHz = getHz();

	initExtensions(hdc);

    CString str;

    infoString = "Relevant OpenGL extensions:\n";
    str.Format("GL_ARB_multitexture:             %s\n", GL_ARB_multitexture? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_ARB_texture_compression:      %s\n", GL_ARB_texture_compression? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_EXT_texture_compression_s3tc: %s\n", GL_EXT_texture_compression_s3tc? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_EXT_texture_env_dot3:         %s\n", GL_EXT_texture_env_dot3? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_ARB_texture_env_combine:      %s\n", GL_ARB_texture_env_combine? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_ARB_texture_env_crossbar:     %s\n", GL_ARB_texture_env_crossbar? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_ARB_texture_cube_map:         %s\n", GL_ARB_texture_cube_map? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_EXT_texture_edge_clamp:       %s\n", GL_EXT_texture_edge_clamp? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_EXT_texture3D:                %s\n", GL_EXT_texture3D? "Supported" : "Not supported");
    infoString += str + '\n';
    str.Format("GL_ATI_fragment_shader:          %s\n", GL_ATI_fragment_shader? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_NV_texture_env_combine4:      %s\n", GL_NV_texture_env_combine4? "Supported" : "Not supported");
    infoString += str;
    str.Format("GL_NV_register_combiners:        %s\n", GL_NV_register_combiners? "Supported" : "Not supported");
    infoString += str;


    CString missing;
    if (!GL_ARB_multitexture)             missing += "GL_ARB_multitexture\n";
//    if (!GL_ARB_texture_compression)      missing += "GL_ARB_texture_compression\n";
//    if (!GL_EXT_texture_compression_s3tc) missing += "GL_EXT_texture_compression_s3tc\n";
    if (!GL_EXT_texture_env_dot3)         missing += "GL_EXT_texture_env_dot3\n";
    if (!GL_ARB_texture_env_combine)      missing += "GL_ARB_texture_env_combine\n";
    if (!GL_EXT_texture_edge_clamp)       missing += "GL_EXT_texture_edge_clamp\n";
    if (!GL_EXT_texture3D)                missing += "GL_EXT_texture3D\n";
    if (!(GL_ARB_texture_env_crossbar || GL_NV_texture_env_combine4)) missing += "GL_ARB_texture_env_crossbar/GL_NV_texture_env_combine4\n";

    if (missing.GetLength() > 0){
        MessageBox("The following required OpenGL extensions isn't supported by the graphic card and driver you're using:\n\n" + missing +
                   "\nThis application will now exit.", "Missing extensions", MB_OK);
        texHandler = NULL;
        PostMessage(WM_CLOSE);
        return 0;
    }


	texHandler = new TextureHandler();

    if (GL_ATI_fragment_shader){

        ShaderID fragmentShader = texHandler->beginFragmentShader();
            // Attenuation map
            glSampleMapATI(GL_REG_0_ATI, GL_TEXTURE0_ARB, GL_SWIZZLE_STR_ATI);
            // Normal map
		    glSampleMapATI(GL_REG_1_ATI, GL_TEXTURE1_ARB, GL_SWIZZLE_STR_ATI);
            // Normalizing cubemap
		    glSampleMapATI(GL_REG_2_ATI, GL_TEXTURE2_ARB, GL_SWIZZLE_STR_ATI);

            glColorFragmentOp2ATI(GL_DOT3_ATI, GL_REG_2_ATI, GL_NONE, GL_NONE, //GL_SATURATE_BIT_ATI,
                                            GL_REG_1_ATI, GL_NONE, GL_BIAS_BIT_ATI | GL_2X_BIT_ATI,
                                            GL_REG_2_ATI, GL_NONE, GL_BIAS_BIT_ATI | GL_2X_BIT_ATI);
            glColorFragmentOp2ATI(GL_MUL_ATI,  GL_REG_2_ATI, GL_NONE, GL_NONE,
                                            GL_REG_2_ATI, GL_NONE, GL_NONE,
                                            GL_REG_2_ATI, GL_NONE, GL_NONE);
            glColorFragmentOp2ATI(GL_MUL_ATI,  GL_REG_2_ATI, GL_NONE, GL_NONE,
                                            GL_REG_2_ATI, GL_NONE, GL_NONE,
                                            GL_REG_2_ATI, GL_NONE, GL_NONE);
            glColorFragmentOp2ATI(GL_MUL_ATI,  GL_REG_2_ATI, GL_NONE, GL_NONE, //GL_2X_BIT_ATI | GL_SATURATE_BIT_ATI,
                                            GL_REG_2_ATI, GL_NONE, GL_NONE,
                                            GL_REG_2_ATI, GL_NONE, GL_NONE);

            glColorFragmentOp2ATI(GL_MUL_ATI,  GL_REG_1_ATI, GL_NONE, GL_NONE,
                                            GL_REG_2_ATI, GL_NONE, GL_NONE,
                                            GL_PRIMARY_COLOR_ARB, GL_NONE, GL_NONE);

            // Maybe unnessesary
            //glAlphaFragmentOp1ATI(GL_MOV_ATI,  GL_REG_1_ATI, GL_NONE, 
            //                                GL_PRIMARY_COLOR_ARB, GL_NONE, GL_NONE);

            glColorFragmentOp2ATI(GL_MUL_ATI,  GL_REG_0_ATI, GL_NONE, GL_SATURATE_BIT_ATI,
                                            GL_REG_0_ATI, GL_NONE, GL_NONE,
                                            GL_REG_1_ATI, GL_NONE, GL_NONE);
        texHandler->endFragmentShader(fragmentShader);
    	

    } else if (GL_NV_register_combiners){
        GLint nGeneralCombiners;
        glGetIntegerv(GL_MAX_GENERAL_COMBINERS_NV, &nGeneralCombiners);

        if (nGeneralCombiners <= 6){
            GL_NV_register_combiners = false;
        } else {
            ShaderID registerCombiner = texHandler->beginRegisterCombiner();

                glCombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV, 6);
                // Normal map
                glCombinerInputNV(GL_COMBINER0_NV,  GL_RGB, GL_VARIABLE_A_NV, GL_TEXTURE1_ARB, GL_EXPAND_NORMAL_NV, GL_RGB);
                // Normalizing cubemap
                glCombinerInputNV(GL_COMBINER0_NV,  GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE2_ARB, GL_EXPAND_NORMAL_NV, GL_RGB);
                glCombinerOutputNV(GL_COMBINER0_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV, GL_NONE, GL_NONE, GL_TRUE, GL_FALSE, GL_FALSE);

                glCombinerInputNV(GL_COMBINER1_NV,  GL_RGB, GL_VARIABLE_A_NV, GL_SPARE0_NV, GL_SIGNED_IDENTITY_NV, GL_RGB);
                glCombinerInputNV(GL_COMBINER1_NV,  GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV, GL_SIGNED_IDENTITY_NV, GL_RGB);
                glCombinerOutputNV(GL_COMBINER1_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);

                glCombinerInputNV(GL_COMBINER2_NV,  GL_RGB, GL_VARIABLE_A_NV, GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glCombinerInputNV(GL_COMBINER2_NV,  GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glCombinerOutputNV(GL_COMBINER2_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);

                glCombinerInputNV(GL_COMBINER3_NV,  GL_RGB, GL_VARIABLE_A_NV, GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glCombinerInputNV(GL_COMBINER3_NV,  GL_RGB, GL_VARIABLE_B_NV, GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glCombinerOutputNV(GL_COMBINER3_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);

                glCombinerInputNV(GL_COMBINER4_NV,  GL_RGB, GL_VARIABLE_A_NV, GL_SPARE0_NV,        GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glCombinerInputNV(GL_COMBINER4_NV,  GL_RGB, GL_VARIABLE_B_NV, GL_PRIMARY_COLOR_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glCombinerOutputNV(GL_COMBINER4_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);

                glCombinerInputNV(GL_COMBINER5_NV,  GL_RGB, GL_VARIABLE_A_NV, GL_SPARE0_NV,    GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                // Attenuation map
                glCombinerInputNV(GL_COMBINER5_NV,  GL_RGB, GL_VARIABLE_B_NV, GL_TEXTURE0_ARB, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glCombinerOutputNV(GL_COMBINER5_NV, GL_RGB, GL_SPARE0_NV, GL_DISCARD_NV, GL_DISCARD_NV, GL_NONE, GL_NONE, GL_FALSE, GL_FALSE, GL_FALSE);

                glFinalCombinerInputNV(GL_VARIABLE_A_NV, GL_ZERO,      GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glFinalCombinerInputNV(GL_VARIABLE_B_NV, GL_ZERO,      GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glFinalCombinerInputNV(GL_VARIABLE_C_NV, GL_ZERO,      GL_UNSIGNED_IDENTITY_NV, GL_RGB);
                glFinalCombinerInputNV(GL_VARIABLE_D_NV, GL_SPARE0_NV, GL_UNSIGNED_IDENTITY_NV, GL_RGB);
            texHandler->endRegisterCombiner(registerCombiner);
        }
    }



	Set <class Polygon *> polys;

    lights.add(new Light(Vertex(0,0,0), Color(1,1,1), 700));

	
	int i, j;
    for (i = 0; i < 8; i++){
	    j = lights.add(new Light(Vertex(420.0f * cosf((i + 0.5f) * 2 * PI / 8), 465, 420.0f * sinf((i + 0.5f) * 2 * PI / 8)), Color(float(!((i >> 1) & i)),float((i >> 1) & 1),float(i & 1)), 250));
		lights[j]->setCorona(Vertex(-cosf((i + 0.5f) * 2 * PI / 8), 0, -sinf((i + 0.5f) * 2 * PI / 8)), 0.5f, false);
    }
    for (i = 0; i < 8; i++){
	    j = lights.add(new Light(Vertex(675.0f * cosf((i + 0.5f) * 2 * PI / 8), 280, 675.0f * sinf((i + 0.5f) * 2 * PI / 8)), Color(1,1,1), 250));
		lights[j]->setCorona(Vertex(cosf((i + 0.5f) * 2 * PI / 8), 0, sinf((i + 0.5f) * 2 * PI / 8)), 1, false);
    }
    for (i = 0; i < 4; i++){
	    j = lights.add(new Light(Vertex(630.0f * cosf(i * 2 * PI / 4), 120, 630.0f * sinf(i * 2 * PI / 4)), Color(1,1,1), 250));
		lights[j]->setCorona(Vertex(0,-1,0), 1.5f, false);
    }    

    lights.add(new Light(Vertex(0,130,-1440), Color(1,1,1), 150));
    lights.add(new Light(Vertex(0,370,-1250), Color(1,1,1), 300));
    lights.add(new Light(Vertex(0,370,-1440), Color(1,1,1), 300));
    lights.add(new Light(Vertex(0,370,-1630), Color(1,1,1), 300));
    lights.add(new Light(Vertex(0,370,-1820), Color(1,1,1), 300));

    lights.add(new Light(Vertex(0,130, 1440), Color(1,1,1), 150));
    lights.add(new Light(Vertex(0,370, 1250), Color(1,1,1), 300));
    lights.add(new Light(Vertex(0,370, 1440), Color(1,1,1), 300));
    lights.add(new Light(Vertex(0,370, 1630), Color(1,1,1), 300));
    lights.add(new Light(Vertex(0,370, 1820), Color(1,1,1), 300));

	for (i = 20; i < 31; i++){
		lights[i]->setCorona(Vertex(0,-1,0), 1, false);
	}

    lights.add(new Light(Vertex(-150,-50, 1440), Color(1,1,1), 300));
    lights.add(new Light(Vertex( 150,-50, 1440), Color(1,1,1), 300));
    lights.add(new Light(Vertex(-150,-50,-1440), Color(1,1,1), 300));
    lights.add(new Light(Vertex( 150,-50,-1440), Color(1,1,1), 300));

    lights.add(new Light(Vertex(0,-50, 1200), Color(1,1,1), 300));
    lights.add(new Light(Vertex(0,-50,-1200), Color(1,1,1), 300));

    TextureID normalizingCubeMap, attenuationMap, vectorField;//, specularMap;
    normalizingCubeMap = texHandler->createNormalizingCubeMap(32);
    attenuationMap     = texHandler->createAttenuationMap(32);
    vectorField        = texHandler->createVectorField(64);
    //specularMap        = texHandler->createSpecularMap(64,64, 0, 40);

	text     = texHandler->loadFromFile("..\\Textures\\Font2.dds");
	lightTex = texHandler->loadFromFile("..\\Textures\\pearl.jpg");
	corona   = texHandler->loadFromFile("..\\Textures\\Corona2.png");


    TextureID wall, wallBump;
	wall      = texHandler->loadFromFile("..\\Textures\\wood.dds");
	wallBump  = texHandler->loadFromFile("..\\Textures\\woodBump.png");
    texHandler->toNormalMap(wallBump);
    texHandler->setInternalFormat(wallBump, GL_RGB5);

    TextureID floor, floorBump;
	floor     = texHandler->loadFromFile("..\\Textures\\Fieldstone.dds");
	floorBump = texHandler->loadFromFile("..\\Textures\\FieldstoneBump.png");
    texHandler->toNormalMap(floorBump);
    texHandler->setInternalFormat(floorBump, GL_RGB5);

    TextureID floor2, floor2Bump;
	floor2     = texHandler->loadFromFile("..\\Textures\\brickfloor.dds");
	floor2Bump = texHandler->loadFromFile("..\\Textures\\brickfloorBump.png");
    texHandler->toNormalMap(floor2Bump);
    texHandler->setInternalFormat(floor2Bump, GL_RGB5);

    TextureID metal, metalBump;
	metal     = texHandler->loadFromFile("..\\Textures\\metal.dds");
	metalBump = texHandler->loadFromFile("..\\Textures\\metalBump.png");
    texHandler->toNormalMap(metalBump);
    texHandler->setInternalFormat(metalBump, GL_RGB5);

    TextureID lite, liteBump;
	lite       = texHandler->loadFromFile("..\\Textures\\rnd_lite1.dds");
	liteBump   = texHandler->loadFromFile("..\\Textures\\rnd_lite1Bump.png");
    texHandler->toNormalMap(liteBump);
    texHandler->setInternalFormat(liteBump, GL_RGB5);


    TextureTable texTable;
    texTable.addTextureName("rClfWl1a",    256, 256, wall,  wallBump,  PF_DIFFUSE);
    texTable.addTextureName("rClfFlr1x",   256, 256, floor2,floor2Bump,(GL_ATI_fragment_shader || GL_NV_register_combiners)? PF_SPECULAR : PF_DIFFUSE);
    texTable.addTextureName("rClfPlr1",    128, 256, metal, metalBump, (GL_ATI_fragment_shader || GL_NV_register_combiners)? PF_SPECULAR : PF_DIFFUSE);
    texTable.addTextureName("rClfFlr6x",   256, 256, floor, floorBump, (GL_ATI_fragment_shader || GL_NV_register_combiners)? PF_SPECULAR : PF_DIFFUSE);
    texTable.addTextureName("rnd_lite1",   64,  128, lite,  liteBump,  PF_DIFFUSE);



	T3dLoader t3dLoader;
	t3dLoader.loadFromFile("testbana2.t3d", polys, texTable);



	bspTree = new BSPTree(polys);

	texHandler->uploadAll();


    demoLooper.loadFromFile("path.pth");


    if (playMusic){
    	BASS_Init(-1,44100,BASS_DEVICE_3D,(HWND) *this);
	    BASS_SetGlobalVolumes(100,100,100);
    	BASS_Start();

        hMusic = BASS_MusicLoad(FALSE, "Aryx.mod", 0, 0, BASS_MUSIC_LOOP);
        BASS_MusicPlay(hMusic);
    } else {
        hMusic = NULL;
    }

	return 0;
}

void OpenGLWindow::OnSize(UINT nType, int cx, int cy){
	glViewport(0, 0, cx, cy);
}

void OpenGLWindow::OnClose(){
    if (playMusic){
        if (hMusic != NULL) BASS_MusicFree(hMusic);
        BASS_Stop();
        BASS_Free();
    }

    if (texHandler != NULL){
        int i, len = lights.getSize();
        for (i = 0; i < len; i++) delete lights[i];

        delete bspTree;
        texHandler->unloadAll();
        delete texHandler;
    }


	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(hrc);
	::ReleaseDC((HWND) *this,hdc);

	if (fullScreen) resetResolution();

	ShowCursor(TRUE);
	DestroyWindow();
}


void OpenGLWindow::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags){
	Key[nChar & 0xFF] = true;

    switch(nChar){
        case ' ':
            showFPS = !showFPS;
            break;
        case 'I':
            showInfo = !showInfo;
            break;
        case 'M':
            invertMouse = !invertMouse;
            break;
        case 'L':
            showLights = !showLights;
            break;
        case 'D':
            if (demoMode){
                demoMode = false;
            } else {
                if (demoLooper.getNumberOfNodes() > 0){
                    demoMode = true;
                    demoTime = 0.0f;
                }
            }
            break;
        case 'O':
            demoLooper.loadFromFile("path.pth");
            break;
        case 'S':
            demoLooper.writeToFile("path.pth");
            break;
		case 'C':
			demoLooper.clear();
			break;
        case VK_INSERT:
            demoLooper.addNode(position.x, position.y, position.z, Wx, Wy, Wz);
            break;
        case VK_ESCAPE:
            if (!fullScreen && captureMouse){
                ShowCursor(TRUE);
                captureMouse = false;
            } else {
                PostMessage(WM_CLOSE, 0,0);
            }
            break;
    }
}

void OpenGLWindow::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags){
	Key[nChar & 0xFF] = false;
}

void OpenGLWindow::OnSysKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags){
	if (((nFlags >> 13) & 0x3) == 1 && nChar == VK_ENTER){
		if (fullScreen){
			resetResolution();
			middleX = (windowedRect.left + windowedRect.right)  / 2;
			middleY = (windowedRect.top  + windowedRect.bottom) / 2;
			MoveWindow(windowedRect, TRUE);

			fullScreen = false;
		} else {
			middleX = Width  / 2;
			middleY = Height / 2;
			MoveWindow(0, 0, Width, Height, TRUE);
			setResolution(Width, Height, Bpp, Refresh);

			captureMouse = true;
			ShowCursor(FALSE);

			fullScreen = true;
		}
	}
}

void OpenGLWindow::OnLButtonDown(UINT nFlags, CPoint point){
	mouseX = point.x;
	mouseY = point.y;

	windowMoved = false;
	leftButton = true;
}

void OpenGLWindow::OnLButtonUp(UINT nFlags, CPoint point){
	if (!captureMouse){
		if (!windowMoved){
			captureMouse = true;
			ShowCursor(FALSE);
		}
	}

	leftButton = false;
}

void OpenGLWindow::OnMouseMove(UINT nFlags, CPoint point){
	static bool changeMouse = false;

	if (captureMouse){
		ClientToScreen(&point);

		Wx += mouseSensibility * (point.y - middleY) * (invertMouse? -1 : 1);
		Wy += mouseSensibility * (point.x - middleX);
		
		if (changeMouse = !changeMouse) SetCursorPos(middleX, middleY);
	} else {
		if (leftButton){
			windowedRect += CSize(point.x - mouseX, point.y - mouseY);
			middleX = (windowedRect.left + windowedRect.right)  / 2;
			middleY = (windowedRect.top  + windowedRect.bottom) / 2;

			MoveWindow(windowedRect, TRUE);
			windowMoved = true;
		}
	}
}

void OpenGLWindow::OnKillFocus(CWnd *pNewWnd){
	if (fullScreen){
		resetResolution();
		middleX = (windowedRect.left + windowedRect.right)  / 2;
		middleY = (windowedRect.top  + windowedRect.bottom) / 2;
		MoveWindow(windowedRect, TRUE);
		fullScreen = false;
		captureMouse = false;
		ShowCursor(TRUE);
	}
}

/* ################################################################################################################### */


class OpenGLApp : public CWinApp {
public:
    virtual BOOL InitInstance();
	BOOL OnIdle(LONG lCount);
};


BOOL OpenGLApp::InitInstance(){
	
/*#ifdef _DEBUG
	int flag = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG); // Get current flag
	flag |= _CRTDBG_LEAK_CHECK_DF; // Turn on leak-checking bit
	flag |= _CRTDBG_CHECK_ALWAYS_DF; // Turn on CrtCheckMemory 
	_CrtSetDbgFlag(flag); // Set flag to the new value
#endif
*/

	m_pMainWnd = new OpenGLWindow();

	return TRUE;
}

BOOL OpenGLApp::OnIdle(LONG lCount){
	
	((OpenGLWindow *) m_pMainWnd)->DrawFrame();

	return TRUE;
}

OpenGLApp openGLApp;


/* ########################################## End of file ############################################################ */
